use std::rc::Rc;

use crate::flow::traversible::*;
use crate::func::*;
use crate::rstd::{fallible::*, *};

use super::*;

pub type TPE<'a, Ctx, A> = TreeParseError<ParseError<'a, Ctx, Fctr<'a, Ctx, A>>>;

pub type TreeFaiure<'a, Ctx, A> =
    ResolutionError<<Ctx as Context<'a>>::LookupError, TPE<'a, Ctx, A>>;

pub type SubsetWrapped<'a, Ctx, A> = FallibleWrapped<'a, Ctx, bool, TreeFaiure<'a, Ctx, A>>;

pub type SubsetMonad<'a, Ctx, A> = FallibleMonad<'a, Ctx, TreeFaiure<'a, Ctx, A>>;

pub fn subset_pure<'a, Ctx: Context<'a>, A: Mentionable<'a, Ctx>>(
    value: bool,
) -> SubsetWrapped<'a, Ctx, A> {
    <SubsetMonad<'a, Ctx, A> as Pure>::pure(value)
}

#[derive(PartialEq)]
enum NodeType {
    Rb,
    R,
    B,
}

type RefData = (NodeType, Hash);

impl<'a, Ctx: Context<'a>, A: Mentionable<'a, Ctx>>
    TraversibleBinaryNode<'a, SubsetMonad<'a, Ctx, A>, A, RefData> for BNode<'a, Ctx, A>
{
    fn split(&self) -> Split<'a, SubsetMonad<'a, Ctx, A>, A, RefData> {
        (
            Rc::new(self.cl.clone()),
            Rc::new(self.cr.clone()),
            self.key.clone(),
        )
    }

    fn to_tree(
        self: Rc<Self>,
    ) -> Rc<dyn TraversibleBinaryTree<'a, SubsetMonad<'a, Ctx, A>, A, RefData>> {
        Rc::new(Nullable::from(self))
    }
}

impl<'a, Ctx: Context<'a>, A: Mentionable<'a, Ctx>>
    TraversibleBinaryNode<'a, SubsetMonad<'a, Ctx, A>, A, RefData> for RNode<'a, Ctx, A>
{
    fn split(&self) -> Split<'a, SubsetMonad<'a, Ctx, A>, A, RefData> {
        (
            Rc::new(self.cl.clone()),
            Rc::new(self.cr.clone()),
            self.key.clone(),
        )
    }

    fn to_tree(
        self: Rc<Self>,
    ) -> Rc<dyn TraversibleBinaryTree<'a, SubsetMonad<'a, Ctx, A>, A, RefData>> {
        Rc::new(Nullable::from(self))
    }
}

impl<'a, Ctx: Context<'a>, A: Mentionable<'a, Ctx>>
    TraversibleBinaryNode<'a, SubsetMonad<'a, Ctx, A>, A, RefData> for RBNode<'a, Ctx, A>
{
    fn split(&self) -> Split<'a, SubsetMonad<'a, Ctx, A>, A, RefData> {
        match self {
            RBNode::R(r) => r.split(),
            RBNode::B(b) => b.split(),
        }
    }

    fn to_tree(
        self: Rc<Self>,
    ) -> Rc<dyn TraversibleBinaryTree<'a, SubsetMonad<'a, Ctx, A>, A, RefData>> {
        Rc::new(Nullable::from(self))
    }
}

impl<'a, Ctx: Context<'a>, A: Mentionable<'a, Ctx>>
    TraversibleBinaryReference<'a, SubsetMonad<'a, Ctx, A>, A, RefData>
    for Point<'a, Ctx, BNode<'a, Ctx, A>>
{
    fn resolve(
        &self,
    ) -> Wrap<
        'a,
        Rc<dyn TraversibleBinaryNode<'a, SubsetMonad<'a, Ctx, A>, A, RefData>>,
        SubsetMonad<'a, Ctx, A>,
    > {
        <SubsetMonad<'a, Ctx, A> as Functor>::fmap(
            |resolved| resolved as Rc<dyn TraversibleBinaryNode<'a, _, _, _>>,
            Ctx::stuff(self.resolve()),
        )
    }

    fn data(&self) -> RefData {
        (NodeType::B, self.point)
    }
}

impl<'a, Ctx: Context<'a>, A: Mentionable<'a, Ctx>>
    TraversibleBinaryReference<'a, SubsetMonad<'a, Ctx, A>, A, RefData>
    for Point<'a, Ctx, RNode<'a, Ctx, A>>
{
    fn resolve(
        &self,
    ) -> Wrap<
        'a,
        Rc<dyn TraversibleBinaryNode<'a, SubsetMonad<'a, Ctx, A>, A, RefData>>,
        SubsetMonad<'a, Ctx, A>,
    > {
        <SubsetMonad<'a, Ctx, A> as Functor>::fmap(
            |resolved| resolved as Rc<dyn TraversibleBinaryNode<'a, _, _, _>>,
            Ctx::stuff(self.resolve()),
        )
    }

    fn data(&self) -> RefData {
        (NodeType::R, self.point)
    }
}

impl<'a, Ctx: Context<'a>, A: Mentionable<'a, Ctx>>
    TraversibleBinaryReference<'a, SubsetMonad<'a, Ctx, A>, A, RefData>
    for Point<'a, Ctx, RBNode<'a, Ctx, A>>
{
    fn resolve(
        &self,
    ) -> Wrap<
        'a,
        Rc<dyn TraversibleBinaryNode<'a, SubsetMonad<'a, Ctx, A>, A, RefData>>,
        SubsetMonad<'a, Ctx, A>,
    > {
        <SubsetMonad<'a, Ctx, A> as Functor>::fmap(
            |resolved| resolved as Rc<dyn TraversibleBinaryNode<'a, _, _, _>>,
            Ctx::stuff(self.resolve()),
        )
    }

    fn data(&self) -> RefData {
        (NodeType::Rb, self.point)
    }
}

impl<'a, Ctx: Context<'a>, A: Mentionable<'a, Ctx>, T: Mentionable<'a, Ctx>, D: 'a + PartialEq>
    TraversibleBinaryTree<'a, SubsetMonad<'a, Ctx, A>, A, D> for Nullable<'a, Ctx, T>
where
    Point<'a, Ctx, T>: TraversibleBinaryReference<'a, SubsetMonad<'a, Ctx, A>, A, D>,
{
    fn refer(
        &self,
    ) -> Option<Rc<dyn TraversibleBinaryReference<'a, SubsetMonad<'a, Ctx, A>, A, D>>> {
        match self {
            Nullable::Null(_) => None,
            Nullable::NotNull(point) => Some(Rc::new(point.clone())),
        }
    }
}
